--Create frame for event registration
local Coylist = CreateFrame("Frame", "Coylist", UIParent)
Coylist:SetScript("OnEvent", function(this, event, ...)	this[event](this, ...) end)
Coylist:RegisterEvent("ADDON_LOADED")
Coylist:RegisterEvent("QUEST_LOG_UPDATE")

--Variables
local addon_version = "1.3"	--not using .toc because it's only reloaded by restarting the client
local function defaultGlobalOptions()
	return {
		version = addon_version,
		reset = {	--settings regarding reset times/days
			weeklyDay = 3,
			useWowDailyTime = true,	--if true, use GetQuestResetTime. if false, uses dailyHour and dailyMinute settings
			dailyHour = 6,	--reset at this hour, if useWowDailyTime is false
			dailyMin = 0,	--reset at this minute, if useWowDailyTime is false,
			checkOnlyOnLoad = true,
			resetCheckInterval = 15 --interval in minutes to check for list resetting
		}
	}
end

local function defaultCharacterOptions()
	return {
		defaultFrame = {	--default frame settings
			x = 400,
			y = 400,
			width = 200,
			height = 300,
			backdrop = {	--basic white texture to allow simple tinting to change the color
				bgFile = "Interface\\BUTTONS\\WHITE8x8.png",
				edgeFile = "None",
				Tile = false,
				TileSize = 32,
				EdgeSize = 8,
				insets = {
					left = 2,
					right = 2,
					top = 2,
					bottom = 2
				}
			},
			backdropColor = {0, 0, 0, .5},
			backdropBorderColor = {0, 0, 0, 0},
			padding = 3,
			listSpacing = 3,
			entrySpacing = 2
		},
		frames = {	--user frame data is in here
			{
				frameId = 1,
				visible = true,
				x = 400,
				y = 400,
				name = "Default Frame"
			}
		},
		lists = {	--user list data is in here
		
		},
		locked = false,
		alwaysShowOnLoad = false,
		showOnReset = true,
		showFrameNames = true,
		collapseCompleteLists = false
	}
end

local _frames = {}	--screen elements
--functions
local BuildOptions, DrawFrames, DrawRow, DrawLists, AddList, AddEntry, GetCheckboxHeight, MigrateOptions, CreateNewFrame, GetFrameById, DrawListToggle, Draw_OptionsLists, Draw_OptionsFrames, ShowColorPicker, AddFrame, CreateFontString, 
ToggleLocked, CalculateNextResetTime, CheckResets, DeepCopy, Coylist_OnUpdate

local listFont = CreateFont("Coylist_ListFont")
listFont:SetFont("Fonts\\FRIZQT__.TTF", 10);listFont:SetTextColor(1,.82,0)
local entryFont = CreateFont("Coylist_EntryFont")
entryFont:SetFont("Fonts\\FRIZQT__.TTF", 10);entryFont:SetTextColor(1,1,1)

local weeklyDay_Options = {
	"Sunday",
	"Monday",
	"Tuesday",
	"Wednesday",
	"Thursday",
	"Friday",
	"Saturday"
}

local listType_Options = {
	"Daily",
	"Weekly",
	"Manual",
	"One Time"
}

--set up options for the add entry dialogue box
StaticPopupDialogs["COYLIST_ADDENTRY"] = {
	text = "Enter the text for the new entry",
	button1 = "Ok",
	button2 = "Cancel",
	timeout = 0,
	whileDead = true,
	hideOnEscape = true,
	hasEditBox = true,
	OnShow = function(self)
		self.button1:Disable()
		self.editBox:SetScript("OnEnterPressed", function() self.button1:Click() end)
	end,
	OnAccept = function(self, list)
		AddEntry(list, self.editBox:GetText())
		DrawLists()
	end,
	EditBoxOnTextChanged = function (self, data)
		if(self:GetText() ~= "") then
			self:GetParent().button1:Enable()
		else
			self:GetParent().button1:Disable()
		end
	end
}

--Register slash command
SLASH_COYLIST1 = "/coylist"
function SlashCmdList.COYLIST(args)
	if args == "reset" then
		Coylist_GlobalData = nil
		Coylist_CharData = nil
	elseif args == "show" then
		for i, v in ipairs(_frames) do
			v.settings.visible = true
			v.panel:Show()
		end
	elseif args == "hide" then
		for i, v in ipairs(_frames) do
			v.settings.visible = false
			v.panel:Hide()
		end
	elseif args == "lock" then
		ToggleLocked()
	else
		--calling it once won't switch to the right panel if the interface options hasn't been opened yet
		InterfaceOptionsFrame_OpenToCategory("Coylist-List Settings")
		InterfaceOptionsFrame_OpenToCategory("Coylist-List Settings")
	end
end

local _timeSinceLastUpdate = 0
function Coylist_OnUpdate(self, elapsed)
	if(not (Coylist_GlobalData.reset.checkOnlyOnLoad)) then
		_timeSinceLastUpdate = _timeSinceLastUpdate + elapsed
		if(_timeSinceLastUpdate > (Coylist_GlobalData.reset.resetCheckInterval * 60)) then
			CheckResets()
			_timeSinceLastUpdate = 0
		end
	end
end

function Coylist:ADDON_LOADED(addonName)
	if addonName ~= "Coylist" then return end
	
	--Load in saved data, using defaults if data does not exist
	Coylist_GlobalData = Coylist_GlobalData or defaultGlobalOptions()
	Coylist_CharData = Coylist_CharData or defaultCharacterOptions()
	
	--This block handles if options are added and the user is on a previous version
	--their settings are preserved when possible
	if Coylist_GlobalData.version ~= addon_version then
		MigrateOptions(Coylist_GlobalData, defaultGlobalOptions())
		MigrateOptions(Coylist_CharData, defaultCharacterOptions())
	end
	
	BuildOptions()	
	
	if(Coylist_CharData.alwaysShowOnLoad) then
		--make sure all frames are visible
		for frames_i, frames_v in ipairs(Coylist_CharData.frames) do
			frames_v.visible = true
		end
	end
	
	DrawFrames()
end

--need to wait until the second QUEST_LOG_UPDATE before checking resets, or the timer could be wrong
local logUpdates = 0
function Coylist:QUEST_LOG_UPDATE()
	logUpdates = logUpdates + 1
	
	if(logUpdates >= 2) then
		Coylist:UnregisterEvent("QUEST_LOG_UPDATE")	--unregister this event handler
		
		if(not Coylist_GlobalData.reset.checkOnlyOnLoad) then
			Coylist:SetScript("OnUpdate", Coylist_OnUpdate)
		end
		CheckResets()
	end
end

local _optionsLists_rows = {}
function Draw_OptionsLists(frame, charSettings)
	for i, v in ipairs(charSettings.lists) do
		local row = nil
		if(i > table.getn(_optionsLists_rows)) then
			row = {}
			row.container = CreateFrame("Frame", nil, frame)
			row.container:SetWidth(2000)
			row.container:SetHeight(20)
			
			row.visibleButton = CreateFrame("Button", nil, row.container, "UIPanelButtonTemplate")
			row.visibleButton:SetPoint("TOPLEFT", 5, 0)
			row.visibleButton:SetWidth(60)
			
			row.editBox = CreateFrame("EditBox", nil, row.container, "InputBoxTemplate")
			row.editBox:SetAutoFocus(false)
			row.editBox:SetPoint("TOPLEFT", 75, 0)
			row.editBox:SetPoint("BOTTOMRIGHT", row.container, "TOPLEFT", 275, -20)
			
			row.dropdown_listType = CreateFrame("Frame", "Coylist_OptionsLists_ListType"..i, row.container, "UIDropDownMenuTemplate")
			row.dropdown_listType:ClearAllPoints()
			row.dropdown_listType:SetPoint("TOPLEFT", 275, 0)				
			UIDropDownMenu_SetWidth(row.dropdown_listType, 80)
			UIDropDownMenu_SetButtonWidth(row.dropdown_listType, 96)	
			UIDropDownMenu_JustifyText(row.dropdown_listType, "LEFT")
			
			row.dropdown_frame = CreateFrame("Frame", "Coylist_OptionsLists_Frame"..i, row.container, "UIDropDownMenuTemplate")
			row.dropdown_frame:ClearAllPoints()
			row.dropdown_frame:SetPoint("TOPLEFT", 385, 0)				
			UIDropDownMenu_SetWidth(row.dropdown_frame, 80)
			UIDropDownMenu_SetButtonWidth(row.dropdown_frame, 96)	
			UIDropDownMenu_JustifyText(row.dropdown_frame, "LEFT")
			
			row.deleteButton = CreateFrame("Button", nil, row.container, "UIPanelButtonTemplate")
			row.deleteButton:SetText("Delete")
			row.deleteButton:SetPoint("TOPLEFT", 510, 0)
			row.deleteButton:SetWidth(60)
						
			table.insert(_optionsLists_rows, row)
		else
			row = _optionsLists_rows[i]
			row.container:Show()
		end
		
		row.container:SetPoint("TOPLEFT", 0, -30*i+20)
		--row.editBoxFrame:SetPoint("TOPLEFT", 55, 0)
		
		if(v.visible) then
			row.visibleButton:SetText("Visible")	
			row.visibleButton.Left:SetTexture("Interface\\Buttons\\UI-Panel-Button-Up")
			row.visibleButton.Middle:SetTexture("Interface\\Buttons\\UI-Panel-Button-Up")
			row.visibleButton.Right:SetTexture("Interface\\Buttons\\UI-Panel-Button-Up")			
		else
			row.visibleButton:SetText("Hidden")
			row.visibleButton.Left:SetTexture("Interface\\Buttons\\UI-Panel-Button-Disabled")
			row.visibleButton.Middle:SetTexture("Interface\\Buttons\\UI-Panel-Button-Disabled")
			row.visibleButton.Right:SetTexture("Interface\\Buttons\\UI-Panel-Button-Disabled")
		end
		
		row.visibleButton:SetScript("OnClick",
			function(self)
				PlaySound("igMainMenuOptionCheckBoxOn")
				v.visible = not v.visible
				if(v.visible) then
					row.visibleButton:SetText("Visible")
					row.visibleButton.Left:SetTexture("Interface\\Buttons\\UI-Panel-Button-Up")
					row.visibleButton.Middle:SetTexture("Interface\\Buttons\\UI-Panel-Button-Up")
					row.visibleButton.Right:SetTexture("Interface\\Buttons\\UI-Panel-Button-Up")
				else
					row.visibleButton:SetText("Hidden")
					row.visibleButton.Left:SetTexture("Interface\\Buttons\\UI-Panel-Button-Disabled")
					row.visibleButton.Middle:SetTexture("Interface\\Buttons\\UI-Panel-Button-Disabled")
					row.visibleButton.Right:SetTexture("Interface\\Buttons\\UI-Panel-Button-Disabled")
				end
			end)
				
		row.editBox:SetText(v.name)
		row.editBox:SetCursorPosition(0)
		
		row.editBox:SetScript("OnTextChanged", 
			function(self, userInput)
				if(userInput) then v.name = self:GetText() end
			end)
			
		UIDropDownMenu_Initialize(row.dropdown_listType,
			function(self)
				local info = UIDropDownMenu_CreateInfo()
				for oi, ov in pairs(listType_Options) do
					info.text = ov
					info.value = ov
					info.func = 
						function(self)
							UIDropDownMenu_SetSelectedValue(row.dropdown_listType, self.value)
							v.listType = self.value
						end
					
					UIDropDownMenu_AddButton(info, level)
				end
				
				UIDropDownMenu_SetSelectedValue(row.dropdown_listType, v.listType)
			end)
			
		UIDropDownMenu_Initialize(row.dropdown_frame,
			function(self)
				local info = UIDropDownMenu_CreateInfo()
				for oi, ov in pairs(charSettings.frames) do
					if(ov.name == nil or ov.name == "") then
						info.text = "<no name "..ov.frameId..">"
					else
						info.text = ov.name
					end
					info.value = ov.frameId
					info.func = 
						function(self)
							UIDropDownMenu_SetSelectedValue(row.dropdown_frame, self.value)
							v.frameId = self.value
						end
					
					UIDropDownMenu_AddButton(info, level)
				end
				
				UIDropDownMenu_SetSelectedValue(row.dropdown_frame, v.frameId)
			end)
			
		row.deleteButton:SetScript("OnClick",
			function(self)
				PlaySound("igMainMenuOptionCheckBoxOn")
				table.remove(charSettings.lists, i)
				Draw_OptionsLists(frame, charSettings)
			end)
	end
	
	frame.scrollBar:SetMinMaxValues(0, math.max(0, table.getn(charSettings.lists)*30-340))
	
	for j = table.getn(charSettings.lists) + 1, table.getn(_optionsLists_rows) do
		_optionsLists_rows[j].container:Hide()
	end
end

local _optionsFrames_rows = {}
function Draw_OptionsFrames(frame, charSettings)
	for i, v in ipairs(charSettings.frames) do
		local row = nil
		if(i > table.getn(_optionsFrames_rows)) then
			row = {}
			row.container = CreateFrame("Frame", nil, frame)
			row.container:SetWidth(2000)
			row.container:SetHeight(20)
			
			row.visibleButton = CreateFrame("Button", nil, row.container, "UIPanelButtonTemplate")
			row.visibleButton:SetPoint("TOPLEFT", 5, 0)
			row.visibleButton:SetWidth(60)
			
			row.editBox = CreateFrame("EditBox", nil, row.container, "InputBoxTemplate")
			row.editBox:SetAutoFocus(false)
			row.editBox:SetPoint("TOPLEFT", 75, 0)
			row.editBox:SetPoint("BOTTOMRIGHT", row.container, "TOPLEFT", 285, -20)
			
			row.colorPickerButton = CreateFrame("Button", nil, row.container, "UIPanelButtonTemplate")
			row.colorPickerButton:SetPoint("TOPLEFT", 300, 0)
			row.colorPickerButton:SetWidth(80)
			row.colorPickerButton:SetText("Pick Color")
			
			row.defaultColorButton = CreateFrame("Button", nil, row.container, "UIPanelButtonTemplate")
			row.defaultColorButton:SetPoint("TOPLEFT", 385, 0)
			row.defaultColorButton:SetWidth(120)
			row.defaultColorButton:SetText("Use Default Color")
								
			row.deleteButton = CreateFrame("Button", nil, row.container, "UIPanelButtonTemplate")
			row.deleteButton:SetText("Delete")
			row.deleteButton:SetPoint("TOPLEFT", 510, 0)
			row.deleteButton:SetWidth(60)
						
			table.insert(_optionsFrames_rows, row)
		else
			row = _optionsFrames_rows[i]
			row.container:Show()
		end
		
		row.container:SetPoint("TOPLEFT", 0, -30*i+20)
		
		if(v.visible) then
			row.visibleButton:SetText("Visible")	
			row.visibleButton.Left:SetTexture("Interface\\Buttons\\UI-Panel-Button-Up")
			row.visibleButton.Middle:SetTexture("Interface\\Buttons\\UI-Panel-Button-Up")
			row.visibleButton.Right:SetTexture("Interface\\Buttons\\UI-Panel-Button-Up")			
		else
			row.visibleButton:SetText("Hidden")
			row.visibleButton.Left:SetTexture("Interface\\Buttons\\UI-Panel-Button-Disabled")
			row.visibleButton.Middle:SetTexture("Interface\\Buttons\\UI-Panel-Button-Disabled")
			row.visibleButton.Right:SetTexture("Interface\\Buttons\\UI-Panel-Button-Disabled")
		end
		
		row.visibleButton:SetScript("OnClick",
			function(self)
				PlaySound("igMainMenuOptionCheckBoxOn")
				v.visible = not v.visible
				if(v.visible) then
					row.visibleButton:SetText("Visible")
					row.visibleButton.Left:SetTexture("Interface\\Buttons\\UI-Panel-Button-Up")
					row.visibleButton.Middle:SetTexture("Interface\\Buttons\\UI-Panel-Button-Up")
					row.visibleButton.Right:SetTexture("Interface\\Buttons\\UI-Panel-Button-Up")
				else
					row.visibleButton:SetText("Hidden")
					row.visibleButton.Left:SetTexture("Interface\\Buttons\\UI-Panel-Button-Disabled")
					row.visibleButton.Middle:SetTexture("Interface\\Buttons\\UI-Panel-Button-Disabled")
					row.visibleButton.Right:SetTexture("Interface\\Buttons\\UI-Panel-Button-Disabled")
				end
			end)
				
		row.editBox:SetText(v.name)
		row.editBox:SetCursorPosition(0)
		
		row.editBox:SetScript("OnTextChanged", 
			function(self, userInput)
				if(userInput) then v.name = self:GetText() end
			end)
			
		row.colorPickerButton:SetScript("OnClick",
			function(self)
				PlaySound("igMainMenuOptionCheckBoxOn")
				local oldR, oldG, oldB, oldA = unpack(v.backdropColor or charSettings.defaultFrame.backdropColor)
				ShowColorPicker(oldR, oldG, oldB, oldA, 
					function(restore)
						local newR, newG, newB, newA;
						if restore then
							--user cancelled, grab the previous color
							newR, newG, newB, newA = unpack(restore);
						else
							--selected color or opacity changed
							newA, newR, newG, newB = OpacitySliderFrame:GetValue(), ColorPickerFrame:GetColorRGB();
						end

						if(v.backdropColor == nil) then
							v.backdropColor = {}
						end
						
						v.backdropColor[1] = newR
						v.backdropColor[2] = newG
						v.backdropColor[3] = newB
						v.backdropColor[4] = newA
						
						row.defaultColorButton:Enable()
					end)
			end)
			
		if(v.backdropColor == nil) then row.defaultColorButton:Disable() end
		row.defaultColorButton:SetScript("OnClick",
			function(self)
				v.backdropColor = nil
				self:Disable()
			end)
			
		row.deleteButton:SetScript("OnClick",
			function(self)
				PlaySound("igMainMenuOptionCheckBoxOn")
				table.remove(charSettings.frames, i)
				Draw_OptionsFrames(frame, charSettings)
			end)
	end
	
	frame.scrollBar:SetMinMaxValues(0, math.max(0, table.getn(charSettings.frames)*30-340))
	
	for j = table.getn(charSettings.frames) + 1, table.getn(_optionsFrames_rows) do
		_optionsFrames_rows[j].container:Hide()
	end
end

function BuildOptions()
	local globalSettings = nil
	local charSettings = nil
	
	--utility functions
	local function Slider_SetEnabled(slider, enabled)
		if enabled then
			BlizzardOptionsPanel_Slider_Enable(slider)
		else
			BlizzardOptionsPanel_Slider_Disable(slider)
		end
	end
	
	--Create the options interface
	--Main panel
	local optionsMain = CreateFrame("Frame")
	optionsMain.name = "Coylist"
		
	local optionsMain_Title = CreateFontString(optionsMain, GameFontNormalLarge, optionsMain.name)
	optionsMain_Title:SetPoint("TOPLEFT", 16, -16)
	
	local optionsMain_Subtitle = CreateFontString(optionsMain, GameFontHighlightSmall,
		"Options on this screen control overall settings, such as reset times. All of these settings are stored globally.")
	optionsMain_Subtitle:SetPoint("TOPLEFT", optionsMain_Title, "BOTTOMLEFT", 0, -8)
	
	local label_weeklyDay = CreateFontString(optionsMain, GameFontNormalSmall, "Weekly Reset Day")
	label_weeklyDay:SetPoint("TOPLEFT", optionsMain_Subtitle, 0, -32)
	
	local dropdown_weeklyDay = CreateFrame("Frame", "Coylist_Options_DropdownWeeklyDay", optionsMain, "UIDropDownMenuTemplate")
	dropdown_weeklyDay:ClearAllPoints()
	dropdown_weeklyDay:SetPoint("TOPLEFT", label_weeklyDay, "TOPRIGHT", 16, 8)
	
	local function dropdown_weeklyDay_OnClick(self)
		UIDropDownMenu_SetSelectedValue(dropdown_weeklyDay, self.value)
		globalSettings.reset.weeklyDay = self:GetID()
	end
	
	local function dropdown_weeklyDay_initialize(self, level)
		local info = UIDropDownMenu_CreateInfo()
		for k, v in pairs(weeklyDay_Options) do
			info.text = v
			info.value = v
			info.func = dropdown_weeklyDay_OnClick
			UIDropDownMenu_AddButton(info, level)
		end
		
		if globalSettings then
			UIDropDownMenu_SetSelectedID(dropdown_weeklyDay, globalSettings.reset.weeklyDay)
		else
			UIDropDownMenu_SetSelectedID(dropdown_weeklyDay, 1)
		end
	end
		
	UIDropDownMenu_SetWidth(dropdown_weeklyDay, 100)
	UIDropDownMenu_SetButtonWidth(dropdown_weeklyDay, 116)	
	UIDropDownMenu_JustifyText(dropdown_weeklyDay, "LEFT")
	
	local label_dailyTime = CreateFontString(optionsMain, GameFontNormalSmall, "Daily Reset Time")
	label_dailyTime:SetPoint("TOPLEFT", label_weeklyDay, "BOTTOMLEFT", 0, -24)
	
	local slider_resetHourMinute = nil --declaring here so it can be referenced in the OnClick
	local checkbutton_useWowDailyReset = CreateFrame("CheckButton", "Coylist_Options_UseWowDailyReset", optionsMain, "InterfaceOptionsCheckButtonTemplate")	
	checkbutton_useWowDailyReset:SetPoint("TOPLEFT", label_dailyTime, "BOTTOMLEFT", 16, -8)
	getglobal(checkbutton_useWowDailyReset:GetName() .. 'Text'):SetJustifyH("LEFT")	
	getglobal(checkbutton_useWowDailyReset:GetName() .. 'Text'):SetText("Use WoW Daily Reset?")
	
	checkbutton_useWowDailyReset:SetScript("OnClick",
		function(self)
			if(self:GetChecked()) then
				globalSettings.reset.useWowDailyTime = true
				Slider_SetEnabled(slider_resetHourMinute, false)
			else
				globalSettings.reset.useWowDailyTime = false
				Slider_SetEnabled(slider_resetHourMinute, true)
			end
		end)
	
	--declared with the checkbox
	slider_resetHourMinute = CreateFrame("Slider", "Coylist_Options_ResetHourMinute", optionsMain, "OptionsSliderTemplate")	
	slider_resetHourMinute:SetWidth(300)
	slider_resetHourMinute:SetMinMaxValues(0, 95)
	slider_resetHourMinute:SetValueStep(1)
	slider_resetHourMinute:SetObeyStepOnDrag(true)
	slider_resetHourMinute:SetPoint("TOPLEFT", checkbutton_useWowDailyReset, "BOTTOMLEFT", 6, -16)
	getglobal(slider_resetHourMinute:GetName() .. 'Text'):SetText("Daily Reset Time")
	getglobal(slider_resetHourMinute:GetName() .. 'Low'):SetText("00:00")
	getglobal(slider_resetHourMinute:GetName() .. 'High'):SetText("23:45")
	
	slider_resetHourMinute:SetScript("OnValueChanged",
		function(self)
			local hour, minute = math.floor(self:GetValue() / 4), self:GetValue() % 4 * 15
			getglobal(self:GetName() .. 'Text'):SetText(string.format("Daily Reset Time (%02d:%02d)", hour, minute))
			globalSettings.reset.dailyHour = hour
			globalSettings.reset.dailyMin = minute
		end
	)
	
	local slider_resetCheckInterval = nil --declaring here so it can be referenced in the OnClick
	local checkbutton_checkResetOnLoad = CreateFrame("CheckButton", "Coylist_Options_CheckResetOnLoad", optionsMain, "InterfaceOptionsCheckButtonTemplate")	
	checkbutton_checkResetOnLoad:SetPoint("TOPLEFT", checkbutton_useWowDailyReset, "BOTTOMLEFT", 0, -48)
	getglobal(checkbutton_checkResetOnLoad:GetName() .. 'Text'):SetJustifyH("LEFT")	
	getglobal(checkbutton_checkResetOnLoad:GetName() .. 'Text'):SetText("Check for reset only on load?")
	
	checkbutton_checkResetOnLoad:SetScript("OnClick",
		function(self)
			if(self:GetChecked()) then
				globalSettings.reset.checkOnlyOnLoad = true
				Slider_SetEnabled(slider_resetCheckInterval, false)
			else
				globalSettings.reset.checkOnlyOnLoad = false
				Slider_SetEnabled(slider_resetCheckInterval, true)
			end
		end
	)
	
	--declared with the checkbox
	slider_resetCheckInterval = CreateFrame("Slider", "Coylist_Options_ResetCheckInterval", optionsMain, "OptionsSliderTemplate")
	slider_resetCheckInterval:SetWidth(300)
	slider_resetCheckInterval:SetMinMaxValues(5, 60)
	slider_resetCheckInterval:SetValueStep(5)
	slider_resetCheckInterval:SetObeyStepOnDrag(true)
	slider_resetCheckInterval:SetPoint("TOPLEFT", checkbutton_checkResetOnLoad, "BOTTOMLEFT", 6, -16)
	getglobal(slider_resetCheckInterval:GetName() .. 'Text'):SetText("Reset Check Interval")
	getglobal(slider_resetCheckInterval:GetName() .. 'Low'):SetText("5")
	getglobal(slider_resetCheckInterval:GetName() .. 'High'):SetText("60")
	
	slider_resetCheckInterval:SetScript("OnValueChanged",
		function(self)
			getglobal(self:GetName() .. 'Text'):SetText("Reset Check Interval ("..self:GetValue().." minutes)")
			globalSettings.reset.resetCheckInterval = self:GetValue()
		end
	)
			
	InterfaceOptions_AddCategory(optionsMain)
	
	--Child panel - frames
	local optionsFrames = CreateFrame("Frame")
	optionsFrames.name = "Coylist-Frame Settings"
	optionsFrames.parent = optionsMain.name
	
	local optionsFrames_Title = CreateFontString(optionsFrames, GameFontNormalLarge, optionsFrames.name)
	optionsFrames_Title:SetPoint("TOPLEFT", 16, -16)
	
	local optionsFrames_Subtitle = CreateFontString(optionsFrames, GameFontHighlightSmall, 
		"Options on this screen control frames and frame settings. All of these settings are stored per character.")
	optionsFrames_Subtitle:SetPoint("TOPLEFT", optionsFrames_Title, "BOTTOMLEFT", 0, -8)
	
	local checkbutton_showFrameNames = CreateFrame("CheckButton", "Coylist_Options_ShowFrameNames", optionsFrames, "InterfaceOptionsCheckButtonTemplate")	
	checkbutton_showFrameNames:SetPoint("TOPLEFT", optionsFrames_Subtitle, "BOTTOMLEFT", 16, -8)
	getglobal(checkbutton_showFrameNames:GetName() .. 'Text'):SetJustifyH("LEFT")	
	getglobal(checkbutton_showFrameNames:GetName() .. 'Text'):SetText("Show Frame Names?")
	
	checkbutton_showFrameNames:SetScript("OnClick",
		function(self)
			charSettings.showFrameNames = self:GetChecked()
		end)
		
	local checkbutton_alwaysShowOnLoad = CreateFrame("CheckButton", "Coylist_Options_AlwaysShowOnLoad", optionsFrames, "InterfaceOptionsCheckButtonTemplate")	
	checkbutton_alwaysShowOnLoad:SetPoint("TOPLEFT", optionsFrames_Subtitle, "BOTTOMLEFT", 16, -32)
	getglobal(checkbutton_alwaysShowOnLoad:GetName() .. 'Text'):SetJustifyH("LEFT")	
	getglobal(checkbutton_alwaysShowOnLoad:GetName() .. 'Text'):SetText("Always Show Frames On Load?")
	
	checkbutton_alwaysShowOnLoad:SetScript("OnClick",
		function(self)
			charSettings.alwaysShowOnLoad = self:GetChecked()
		end)
		
	local checkbutton_showOnReset = CreateFrame("CheckButton", "Coylist_Options_ShowOnReset", optionsFrames, "InterfaceOptionsCheckButtonTemplate")	
	checkbutton_showOnReset:SetPoint("TOPLEFT", optionsFrames_Subtitle, "BOTTOMLEFT", 16, -56)
	getglobal(checkbutton_showOnReset:GetName() .. 'Text'):SetJustifyH("LEFT")	
	getglobal(checkbutton_showOnReset:GetName() .. 'Text'):SetText("Show Frames On Reset?")
	
	checkbutton_showOnReset:SetScript("OnClick",
		function(self)
			charSettings.showOnReset = self:GetChecked()
		end)
	
	local optionsFrames_Table1Offset = 70
	local optionsFrames_Table2Offset = 300
	
	local optionsFrames_Table1 = CreateFontString(optionsFrames, GameFontNormalSmall, "Frame Name")
	optionsFrames_Table1:SetPoint("TOPLEFT", optionsFrames_Subtitle, "BOTTOMLEFT", optionsFrames_Table1Offset, -96)
	
	local optionsFrames_Table2 = CreateFontString(optionsFrames, GameFontNormalSmall, "Background Color")
	optionsFrames_Table2:SetPoint("TOPLEFT", optionsFrames_Subtitle, "BOTTOMLEFT", optionsFrames_Table2Offset, -96)
	
	local optionsFrames_scrollFrame = CreateFrame("ScrollFrame", nil, optionsFrames)
	optionsFrames_scrollFrame:SetPoint("TOPLEFT", optionsFrames_Subtitle, "BOTTOMLEFT", 0, -112)
	optionsFrames_scrollFrame:SetPoint("RIGHT", optionsFrames, "BOTTOMRIGHT", -16, 0)
	optionsFrames_scrollFrame:SetHeight(350)
	optionsFrames_scrollFrame:EnableMouseWheel(true)
	local optionsFrames_scrollFrameTexture = optionsFrames_scrollFrame:CreateTexture()
	optionsFrames_scrollFrameTexture:SetAllPoints(optionsFrames_scrollFrame)
	optionsFrames_scrollFrameTexture:SetTexture(0, 0, 0, .6)
	
	local optionsFrames_scrollBar = CreateFrame("Slider", "Coylist_ScrollBar_OptionsFrames", optionsFrames_scrollFrame, "UIPanelScrollBarTemplate")
	optionsFrames_scrollBar:SetPoint("TOPRIGHT", 0, -16)
	optionsFrames_scrollBar:SetPoint("BOTTOMRIGHT", 0, 16)
	optionsFrames_scrollBar:SetMinMaxValues(0, 10)
	optionsFrames_scrollBar:SetValueStep(1)
	optionsFrames_scrollBar:SetValue(0)
	optionsFrames_scrollBar:SetWidth(16)
	
	optionsFrames_scrollFrame:SetScript("OnMouseWheel", 
		function(f, value)
			optionsFrames_scrollBar:SetValue(optionsFrames_scrollBar:GetValue() - (value*30))
		end)
	
	local optionsFrames_scrollContent = CreateFrame("Frame", nil, optionsFrames_scrollFrame)
	optionsFrames_scrollContent:SetWidth(1000)
	optionsFrames_scrollContent:SetHeight(1000)
	optionsFrames_scrollFrame:SetScrollChild(optionsFrames_scrollContent)
	optionsFrames_scrollContent.scrollBar = optionsFrames_scrollBar
	
	local optionsFrames_addFrameButton = CreateFrame("Button", nil, optionsFrames, "UIPanelButtonTemplate")
	optionsFrames_addFrameButton:SetPoint("TOPLEFT", optionsFrames_scrollFrame, "BOTTOMLEFT", 0, -16)
	optionsFrames_addFrameButton:SetText("Add Frame")
	optionsFrames_addFrameButton:SetWidth(80)
	
	optionsFrames_addFrameButton:SetScript("OnClick",
		function(self)
			PlaySound("igMainMenuOptionCheckBoxOn")
			AddFrame(charSettings.frames, "New Frame")
			Draw_OptionsFrames(optionsFrames_scrollContent, charSettings)
		end)
		
	InterfaceOptions_AddCategory(optionsFrames)
	
	--Child panel - lists
	local optionsLists = CreateFrame("Frame")
	optionsLists.name = "Coylist-List Settings"
	optionsLists.parent = optionsMain.name
		
	local optionsLists_Title = CreateFontString(optionsLists, GameFontNormalLarge, optionsLists.name)
	optionsLists_Title:SetPoint("TOPLEFT", 16, -16)
	
	local optionsLists_Subtitle = CreateFontString(optionsLists, GameFontHighlightSmall, 
		"Options on this screen control lists and list settings. All of these settings are stored per character.")
	optionsLists_Subtitle:SetPoint("TOPLEFT", optionsLists_Title, "BOTTOMLEFT", 0, -8)
	
	local checkbutton_collapseCompleteLists = CreateFrame("CheckButton", "Coylist_Options_CollapseCompleteLists", optionsLists, "InterfaceOptionsCheckButtonTemplate")	
	checkbutton_collapseCompleteLists:SetPoint("TOPLEFT", optionsLists_Subtitle, "BOTTOMLEFT", 16, -8)
	getglobal(checkbutton_collapseCompleteLists:GetName() .. 'Text'):SetJustifyH("LEFT")	
	getglobal(checkbutton_collapseCompleteLists:GetName() .. 'Text'):SetText("Collapse completed lists?")
	
	checkbutton_collapseCompleteLists:SetScript("OnClick",
		function(self)
			charSettings.collapseCompleteLists = self:GetChecked()
		end)
	
	local optionsLists_Table1Offset = 70
	local optionsLists_Table2Offset = 295
	local optionsLists_Table3Offset = 405
	
	local optionsLists_Table1 = CreateFontString(optionsLists, GameFontNormalSmall, "List Name")
	optionsLists_Table1:SetPoint("TOPLEFT", optionsLists_Subtitle, "BOTTOMLEFT", optionsLists_Table1Offset, -48)
	
	local optionsLists_Table2 = CreateFontString(optionsLists, GameFontNormalSmall, "List Type")
	optionsLists_Table2:SetPoint("TOPLEFT", optionsLists_Subtitle, "BOTTOMLEFT", optionsLists_Table2Offset, -48)
	
	local optionsLists_Table3 = CreateFontString(optionsLists, GameFontNormalSmall, "Frame")
	optionsLists_Table3:SetPoint("TOPLEFT", optionsLists_Subtitle, "BOTTOMLEFT", optionsLists_Table3Offset, -48)
	
	local optionsLists_scrollFrame = CreateFrame("ScrollFrame", nil, optionsLists)
	optionsLists_scrollFrame:SetPoint("TOPLEFT", optionsLists_Subtitle, "BOTTOMLEFT", 0, -64)
	optionsLists_scrollFrame:SetPoint("RIGHT", optionsLists, "BOTTOMRIGHT", -16, 0)
	optionsLists_scrollFrame:SetHeight(350)
	optionsLists_scrollFrame:EnableMouseWheel(true)
	local optionsLists_scrollFrameTexture = optionsLists_scrollFrame:CreateTexture()
	optionsLists_scrollFrameTexture:SetAllPoints(optionsLists_scrollFrame)
	optionsLists_scrollFrameTexture:SetTexture(0, 0, 0, .6)
	
	local optionsLists_scrollBar = CreateFrame("Slider", "Coylist_ScrollBar_OptionsLists", optionsLists_scrollFrame, "UIPanelScrollBarTemplate")
	optionsLists_scrollBar:SetPoint("TOPRIGHT", 0, -16)
	optionsLists_scrollBar:SetPoint("BOTTOMRIGHT", 0, 16)
	optionsLists_scrollBar:SetMinMaxValues(0, 10)
	optionsLists_scrollBar:SetValueStep(1)
	optionsLists_scrollBar:SetValue(0)
	optionsLists_scrollBar:SetWidth(16)
	
	optionsLists_scrollFrame:SetScript("OnMouseWheel", 
		function(f, value)
			optionsLists_scrollBar:SetValue(optionsLists_scrollBar:GetValue() - (value*30))
		end)
	
	local optionsLists_scrollContent = CreateFrame("Frame", nil, optionsLists_scrollFrame)
	optionsLists_scrollContent:SetWidth(1000)
	optionsLists_scrollContent:SetHeight(1000)
	optionsLists_scrollFrame:SetScrollChild(optionsLists_scrollContent)
	optionsLists_scrollContent.scrollBar = optionsLists_scrollBar
	
	local optionsLists_addListButton = CreateFrame("Button", nil, optionsLists, "UIPanelButtonTemplate")
	optionsLists_addListButton:SetPoint("TOPLEFT", optionsLists_scrollFrame, "BOTTOMLEFT", 0, -16)
	optionsLists_addListButton:SetText("Add List")
	optionsLists_addListButton:SetWidth(70)
	
	optionsLists_addListButton:SetScript("OnClick",
		function(self)
			PlaySound("igMainMenuOptionCheckBoxOn")
			AddList(charSettings.lists, "Daily", "New List", charSettings.frames[1].frameId)
			Draw_OptionsLists(optionsLists_scrollContent, charSettings)
		end)
			
	InterfaceOptions_AddCategory(optionsLists)
			
	--these events get fired no matter where in interface options you are, so only need to hook them once
	--don't need to hook cancel, just do nothing
	optionsMain.refresh = (function()
		--copy saved variables into temp settings
		globalSettings = DeepCopy(Coylist_GlobalData)
		charSettings = DeepCopy(Coylist_CharData)
				
		--set up ui
		UIDropDownMenu_Initialize(dropdown_weeklyDay, dropdown_weeklyDay_initialize)
		
		checkbutton_useWowDailyReset:SetChecked(globalSettings.reset.useWowDailyTime)
		
		Slider_SetEnabled(slider_resetHourMinute, not globalSettings.reset.useWowDailyTime)
		slider_resetHourMinute:SetValue((globalSettings.reset.dailyHour * 4) + (globalSettings.reset.dailyMin / 15))
		
		checkbutton_checkResetOnLoad:SetChecked(globalSettings.reset.checkOnlyOnLoad)
		
		Slider_SetEnabled(slider_resetCheckInterval, not globalSettings.reset.checkOnlyOnLoad)
		slider_resetCheckInterval:SetValue(globalSettings.reset.resetCheckInterval)
		
		checkbutton_showFrameNames:SetChecked(charSettings.showFrameNames)
		checkbutton_alwaysShowOnLoad:SetChecked(charSettings.alwaysShowOnLoad)
		checkbutton_showOnReset:SetChecked(charSettings.showOnReset)
		
		Draw_OptionsFrames(optionsFrames_scrollContent, charSettings)
		
		checkbutton_collapseCompleteLists:SetChecked(charSettings.collapseCompleteLists)
		
		Draw_OptionsLists(optionsLists_scrollContent, charSettings)
	end)
	
	optionsMain.okay = (function()
		--copy temp settings back into the saved variables
		Coylist_GlobalData = DeepCopy(globalSettings)
		Coylist_CharData = DeepCopy(charSettings)
		
		if(Coylist_GlobalData.reset.checkOnlyOnLoad) then
			Coylist:SetScript("OnUpdate", nil)
		else
			Coylist:SetScript("OnUpdate", Coylist_OnUpdate)
			CheckResets()
		end
		
		for i, v in ipairs(Coylist_CharData.lists) do
			v.nextReset = CalculateNextResetTime(v.listType)
		end
		
		DrawFrames()
	end)	
end

function DrawRow(frame, index, x, y, item, controlButtonCallback, checkButtonCallback)
	local padding = frame.settings.padding or Coylist_CharData.defaultFrame.padding
	local width = frame.settings.width or Coylist_CharData.defaultFrame.width
	
	--reuse elements when able, otherwise create new ones
	local checkbox = nil
	local controlButton = nil
	if(index > table.getn(frame.checkboxes)) then
		checkbox = CreateFrame("CheckButton", "Coylist_Checkbox_"..frame.settings.frameId.."_"..index, frame.scrollContent, "UICheckButtonTemplate")
		checkbox:SetSize(20, 20)
		getglobal(checkbox:GetName() .. 'Text'):SetJustifyH("LEFT")	
		
		table.insert(frame.checkboxes, checkbox)
		
		controlButton = CreateFrame("Button", "Coylist_ControlButton_"..frame.settings.frameId.."_"..index, frame.scrollContent)
		controlButton:SetSize(15, 15)
		
		table.insert(frame.controlButtons, controlButton)
	else
		checkbox = frame.checkboxes[index]
		controlButton = frame.controlButtons[index]
	end
	
	checkbox:SetPoint("TOPLEFT", x, y)
	checkbox:SetChecked(item.checked)
	checkbox:Show()	--checkbox may be reused and was hidden before, make sure it's visible
	
	--width of the scrollbar and padding
	controlButton:SetPoint("TOPRIGHT", -20-padding, y-1)
	
	controlButton:SetHighlightTexture("Interface/BUTTONS/UI-PlusButton-Hilight.png", "ADD")
	if(item.entries ~= nil) then
		controlButton:SetNormalTexture("Interface/BUTTONS/UI-PlusButton-Up.png")
		controlButton:SetPushedTexture("Interface/BUTTONS/UI-PlusButton-Down.png")
	else
		controlButton:SetNormalTexture("Interface/BUTTONS/UI-MinusButton-Up.png")
		controlButton:SetPushedTexture("Interface/BUTTONS/UI-MinusButton-Down.png")
	end		
	if(Coylist_CharData.locked) then controlButton:Hide() else controlButton:Show() end

	getglobal(checkbox:GetName() .. 'Text'):SetText(item.name)
	
	--width of the window minus: right padding, x position, the width of the checkbox itself, the space between the
	--checkbox and text, the space between the text and control button, the scrollbar width, and the add/remove button
	getglobal(checkbox:GetName() .. 'Text'):SetWidth(width-padding-x-checkbox:GetWidth()-2-2-20-20)
	getglobal(checkbox:GetName() .. 'Text'):SetPoint("TOPLEFT", checkbox, "TOPRIGHT", 2, -3)
	
	checkbox:SetScript("OnClick",
		function(self)
			checkButtonCallback(self:GetChecked())
		end
	)
	controlButton:SetScript("OnClick", 
		function()
			controlButtonCallback()
			PlaySound("igMainMenuOptionCheckBoxOn")
		end)
	
	if(item.entries ~= nil) then
		--list
		getglobal(checkbox:GetName() .. 'Text'):SetFontObject(listFont)
	else
		--entry
		getglobal(checkbox:GetName() .. 'Text'):SetFontObject(entryFont)
	end
	
	return checkbox
end

function DrawListToggle(frame, index, x, y, item)
	local padding = frame.settings.padding or Coylist_CharData.defaultFrame.padding
	
	local toggle = nil
	if(index > table.getn(frame.toggleButtons)) then
		toggle = CreateFrame("Button", "Coylist_Toggle_"..frame.settings.frameId.."_"..index, frame.scrollContent)
		toggle:SetSize(20, 20)
		
		table.insert(frame.toggleButtons, toggle)
	else
		toggle = frame.toggleButtons[index]
	end
	
	toggle:SetHighlightTexture("Interface/BUTTONS/UI-PlusButton-Hilight.png", "ADD")
	
	toggle.SetExpanded = function(self, expanded)
		toggle.expanded = expanded
		if expanded then
			--change appearance
			toggle:SetNormalTexture("Interface/BUTTONS/UI-Panel-CollapseButton-Up.png")
			toggle:SetPushedTexture("Interface/BUTTONS/UI-Panel-CollapseButton-Down.png")
		else
			--change appearance
			toggle:SetNormalTexture("Interface/BUTTONS/UI-Panel-ExpandButton-Up.png")
			toggle:SetPushedTexture("Interface/BUTTONS/UI-Panel-ExpandButton-Down.png")			
		end
	end
	
	toggle:SetScript("OnClick",
		function(self)
			self:SetExpanded(not self.expanded)
			if(self.expanded) then
				item.expanded = true
				PlaySound("igMainMenuOptionCheckBoxOn")
			else
				item.expanded = false
				PlaySound("igMainMenuOptionCheckBoxOff")
			end
			
			DrawLists()
		end
	)
	
	toggle:SetPoint("TOPLEFT", x, y)
	toggle:SetExpanded(item.expanded)
	toggle:Show()
	
	return toggle
end

function DrawLists()
	local frameCheckboxIndexes = {}	--keep track of which checkbox is next for each frame
	local frameToggleIndexes = {}	--keep track of which toggle button is next for each frame
	local frameOffsets = {} --keep track of the current yOffset for each frame
	
	--loop through lists
	for list_i, list_v in ipairs(Coylist_CharData.lists) do
		if(list_v.visible) then
			local frameId = list_v.frameId
			
			--find the frame for this list
			local frame = GetFrameById(frameId)
			
			--don't draw the list if it's set to a deleted frame
			if(frame ~= nil) then
				local padding = frame.settings.padding or Coylist_CharData.defaultFrame.padding
				local listSpacing = frame.settings.listSpacing or Coylist_CharData.defaultFrame.listSpacing
				local entrySpacing = frame.settings.entrySpacing or Coylist_CharData.defaultFrame.entrySpacing
				
				if(frameCheckboxIndexes[frameId] == nil) then frameCheckboxIndexes[frameId] = 1 end
				if(frameToggleIndexes[frameId] == nil) then frameToggleIndexes[frameId] = 1 end
				if(frameOffsets[frameId] == nil) then frameOffsets[frameId] = padding*-1 end
				
				--create the list's checkbox and toggle button
				if(frameCheckboxIndexes[frameId] > 1) then
					frameOffsets[frameId] = frameOffsets[frameId] - (GetCheckboxHeight(frame.checkboxes[frameCheckboxIndexes[frameId]-1]) + listSpacing)
				end
				
				local lastItem = DrawListToggle(frame, frameToggleIndexes[frameId], padding, frameOffsets[frameId], list_v)
				lastItem = DrawRow(frame, frameCheckboxIndexes[frameId], padding+lastItem:GetWidth()-2, frameOffsets[frameId], list_v, 
					function()
						local dialog = StaticPopup_Show ("COYLIST_ADDENTRY")
						dialog.data = list_v
					end,
					function(checked)
						list_v.checked = checked
						
						if(checked) then							
							PlaySound("igMainMenuOptionCheckBoxOn")
							
							if(list_v.listType == "One Time") then
								table.remove(Coylist_CharData.lists, list_i)
							else					
								for listentry_i, listentry_v in ipairs(list_v.entries) do
									listentry_v.checked = true
								end
								
								if(Coylist_CharData.collapseCompleteLists) then list_v.expanded = false end
							end
							DrawLists()
						else
							PlaySound("igMainMenuOptionCheckBoxOff")
							
							for listentry_i, listentry_v in ipairs(list_v.entries) do
								listentry_v.checked = false
							end			
							
							DrawLists()
						end
					end
				)		
				frameCheckboxIndexes[frameId] = frameCheckboxIndexes[frameId] + 1
				frameToggleIndexes[frameId] = frameToggleIndexes[frameId] + 1		
				
				if(list_v.expanded) then
					--create the entry checkboxes
					for entry_i, entry_v in ipairs(list_v.entries) do
						frameOffsets[frameId] = frameOffsets[frameId] - (GetCheckboxHeight(lastItem) + entrySpacing)
						
						lastItem = DrawRow(frame, frameCheckboxIndexes[frameId], padding+26, frameOffsets[frameId], entry_v,
							function()
								table.remove(list_v.entries, entry_i)
								DrawLists()
							end,
							function(checked)
								entry_v.checked = checked
								
								if(checked) then							
									PlaySound("igMainMenuOptionCheckBoxOn")
									
									local checkParent = true
									for sentry_i, sentry_v in ipairs(list_v.entries) do
										if(sentry_v.checked == false) then checkParent = false end
									end
									
									if(checkParent) then
										if(list_v.listType == "One Time") then
											table.remove(Coylist_CharData.lists, list_i)
										else
											list_v.checked = true
											if(Coylist_CharData.collapseCompleteLists) then list_v.expanded = false end
										end
										
										DrawLists()
									end
								else
									PlaySound("igMainMenuOptionCheckBoxOff")
									
									if(list_v.checked == true) then
										list_v.checked = false
									end
									DrawLists()
								end
							end)
						frameCheckboxIndexes[frameId] = frameCheckboxIndexes[frameId] + 1
					end
				end
			end
		end
	end
	
	--for each frame that was updated..
	for k, v in pairs(frameCheckboxIndexes) do
		local frame = GetFrameById(k)
		
		--update the scrollbar height
		local scroll = math.max(-frameOffsets[k]-frame.scrollFrame:GetHeight()+GetCheckboxHeight(frame.checkboxes[frameCheckboxIndexes[k]-1]), 0)
		frame.scrollBar:SetMinMaxValues(0, scroll)
		if(scroll == 0) then frame.scrollBar:Hide() else frame.scrollBar:Show() end
		
		--..hide any extra checkboxes and control buttons
		for i = frameCheckboxIndexes[k], table.getn(frame.checkboxes) do
			frame.checkboxes[i]:Hide()
			frame.controlButtons[i]:Hide()
		end
		
		--..hide any extra toggle buttons
		for i = frameToggleIndexes[k], table.getn(frame.toggleButtons) do
			frame.toggleButtons[i]:Hide()
		end
	end
end

function CreateNewFrame(frameSettings)
	--first, see if there is a frame available to reuse
	local frame = GetFrameById(nil)
	
	if(frame == nil) then
		--no frame was available, so create a new one
		local panel = CreateFrame("Frame", nil, UIParent)
		panel:EnableMouse(true)
		
		--dragging
		panel:SetMovable(true)		
		panel:SetScript("OnMouseDown", function(self, button)
			if not Coylist_CharData.locked and button == "LeftButton" and not self.isMoving then
				self:StartMoving();
				self.isMoving = true;
			end
		end)
		panel:SetScript("OnMouseUp", function(self, button)
			if not Coylist_CharData.locked and button == "LeftButton" and self.isMoving then
				self:StopMovingOrSizing();
				self.isMoving = false;
				frame.settings.x, frame.settings.y = self:GetCenter()
			end
		end)
		
		--resizing
		panel:SetResizable(true)
		panel:SetMinResize(150, 100)
		local resizer = CreateFrame("Button", nil, panel)
		resizer:SetNormalTexture("Interface\\AddOns\\Coylist\\ResizeGrip")
		resizer:SetHighlightTexture("Interface\\AddOns\\Coylist\\ResizeGrip")
		resizer:SetWidth(16)
		resizer:SetHeight(16)
		resizer:SetPoint("BOTTOMRIGHT", 0, 0)
		resizer:SetScript("OnMouseDown", function(self, button) 
			if not Coylist_CharData.locked and button == "LeftButton" and not self.isMoving then
				panel:StartSizing("BOTTOMRIGHT") 
				self.isMoving = true
			end
		end)
		resizer:SetScript("OnMouseUp", function(self, button)
			if not Coylist_CharData.locked and button == "LeftButton" and self.isMoving then
				panel:StopMovingOrSizing()
				self.isMoving = false;
				frame.settings.x, frame.settings.y = panel:GetCenter()
				frame.settings.width, frame.settings.height = panel:GetWidth(), panel:GetHeight()
				frame.scrollContent:SetWidth(panel:GetWidth())
				DrawLists()
			end
		end)
		if(Coylist_CharData.locked) then resizer:Hide() end
		
		--title
		local title = CreateFontString(panel, "GameFontNormal", frameSettings.name)
		title:SetPoint("TOPLEFT", 8, -8)
			
		local hideButton = CreateFrame("Button", nil, panel)
		hideButton:SetSize(25, 25)
		hideButton:SetPoint("TOPRIGHT", -2, -2)
		hideButton:SetNormalTexture("Interface/BUTTONS/UI-Panel-MinimizeButton-Up.png")
		hideButton:SetPushedTexture("Interface/BUTTONS/UI-Panel-MinimizeButton-Down.png")
		hideButton:SetHighlightTexture("Interface/BUTTONS/UI-PlusButton-Hilight.png", "ADD")
		
		hideButton:SetScript("OnClick",
			function(self)
				frame.settings.visible = false
				DrawFrames()
			end)
		
		local scrollFrame = CreateFrame("ScrollFrame", nil, panel)
		scrollFrame:SetPoint("TOPLEFT", title, "BOTTOMLEFT", -6, -5)
		scrollFrame:SetPoint("BOTTOMRIGHT", 0, frameSettings.padding or Coylist_CharData.defaultFrame.padding)
		scrollFrame:EnableMouseWheel(true)		
		
		local scrollBar = CreateFrame("Slider", "Coylist_ScrollBar_"..frameSettings.frameId, scrollFrame, "UIPanelScrollBarTemplate")
		scrollBar:SetPoint("TOPLEFT", scrollFrame, "TOPRIGHT", -20, -16)
		scrollBar:SetPoint("BOTTOMLEFT", scrollFrame, "BOTTOMRIGHT", -20, 24)
		scrollBar:SetMinMaxValues(0, 10)
		scrollBar:SetValueStep(1)
		scrollBar:SetValue(0)
		scrollBar:SetWidth(16)
		
		scrollFrame:SetScript("OnMouseWheel", 
			function(f, value)
				scrollBar:SetValue(scrollBar:GetValue() - (value*30))
			end)
		
		local scrollContent = CreateFrame("Frame", nil, scrollFrame)
		scrollContent:SetWidth(frameSettings.width or Coylist_CharData.defaultFrame.width)
		scrollContent:SetHeight(10)
		scrollFrame:SetScrollChild(scrollContent)
		
		frame = {
			panel = panel,
			title = title,
			resizer = resizer,
			scrollFrame = scrollFrame,
			scrollContent = scrollContent,
			scrollBar = scrollBar,
			checkboxes = {},
			toggleButtons = {},
			controlButtons = {},
			settings = frameSettings
		}
		table.insert(_frames, frame)
	end
	
	--update the frame according to the frame settings
	local panel = frame.panel		
	panel:SetWidth(frameSettings.width or Coylist_CharData.defaultFrame.width);
	panel:SetHeight(frameSettings.height or Coylist_CharData.defaultFrame.height);
	panel:SetPoint("CENTER", nil, "BOTTOMLEFT",
		frameSettings.x or Coylist_CharData.defaultFrame.x,
		frameSettings.y or Coylist_CharData.defaultFrame.y)
	panel:SetBackdrop(frameSettings.backdrop or Coylist_CharData.defaultFrame.backdrop);	
	panel:SetBackdropBorderColor(unpack(frameSettings.backdropBorderColor or Coylist_CharData.defaultFrame.backdropBorderColor));
	
	panel:Hide()	
	return frame
end

function DrawFrames()
	--Hide all panels by default, in case any settings were deleted
	for frames_i, frames_v in ipairs(_frames) do
		frames_v.panel:Hide()
	end
	
	--step over all saved frame settings
	for settings_i, settings_v in ipairs(Coylist_CharData.frames) do
		--find the frame for this frameSettings, if there is one
		local frame = GetFrameById(settings_v.frameId)
		--no frame found, create a new one
		if(frame == nil) then
			frame = CreateNewFrame(settings_v)
		else
			--make sure the frame is pointing to the correct settings object
			frame.settings = settings_v
		end
		
		if(settings_v.visible) then 
			frame.panel:Show() 
		end
		
		if(Coylist_CharData.showFrameNames) then
			frame.title:Show()
		else
			frame.title:Hide()
		end
		
		frame.title:SetText(settings_v.name)
		frame.panel:SetBackdropColor(unpack(settings_v.backdropColor or Coylist_CharData.defaultFrame.backdropColor));
	end
	
	DrawLists()
end

function CalculateNextResetTime(listType)
	local nextDailyResetTime = nil	--time value in seconds
	if(Coylist_GlobalData.reset.useWowDailyTime) then
		--now + time until the next reset
		local currentTime = time()
		local questResetTime = GetQuestResetTime()
		nextDailyResetTime = currentTime + questResetTime
		
		if(nextDailyResetTime == 0) then print("please tell mike you just saw this "..currentTime.." "..questResetTime) end
	else
		--configured hour/minute, +1 day if the time is < now
		local dateTable = date("*t")
		if(dateTable.hour >= Coylist_GlobalData.reset.dailyHour and dateTable.min >= Coylist_GlobalData.reset.dailyMin) then
			--we've passed the reset time for today, so add one day
			dateTable.day = dateTable.day + 1
		end
		--set the date table to match the configured reset time
		dateTable.hour = Coylist_GlobalData.reset.dailyHour
		dateTable.min = Coylist_GlobalData.reset.dailyMin
		dateTable.sec = 0
		--convert to seconds
		nextDailyResetTime = time(dateTable)
	end
	
	if(listType == "Daily") then
		return nextDailyResetTime
	elseif(listType == "Weekly") then
		--shift X number of days until the time is found for the configured weekly day
		--start with the date table for the reset time as found above
		local dateTable = date("*t", nextDailyResetTime)
		
		--determine how many days to shift
		local shift = 0
		if(dateTable.wday <= Coylist_GlobalData.reset.weeklyDay) then
			shift = Coylist_GlobalData.reset.weeklyDay - dateTable.wday
		else
			--remaining days for this week, plus needed days for next week
			shift = 7 - dateTable.wday + Coylist_GlobalData.reset.weeklyDay
		end
		dateTable.day = dateTable.day + shift
		
		--convert to seconds
		return time(dateTable)
	else
		return nil
	end
end

function CheckResets()
	local now = time()
	local dirty = false
	for list_i, list_v in ipairs(Coylist_CharData.lists) do
		if(list_v.nextReset ~= nil and list_v.nextReset <= now) then
			dirty = true
			list_v.nextReset = CalculateNextResetTime(list_v.listType)
			
			list_v.checked = false
			for entry_i, entry_v in ipairs(list_v.entries) do
				entry_v.checked = false
			end
		end
	end
	
	--redraw only if any list was reset
	if(dirty) then 
		if(Coylist_CharData.showOnReset) then
			--make sure all frames are visible
			for frames_i, frames_v in ipairs(Coylist_CharData.frames) do
				frames_v.visible = true
			end
			
			DrawFrames()
		else
			DrawLists()
		end
	end
end

function AddFrame(frames, name)
	local newFrameId = 1
	
	for i, v in ipairs(frames) do
		if(v.frameId >= newFrameId) then
			newFrameId = v.frameId + 1
		end
	end
	
	local newFrame = {
		frameId = newFrameId,
		visible = true,
		x = 400,
		y = 400,
		name = name
	}
	
	table.insert(frames, newFrame)
	
	return newFrame
end

function AddList(lists, listType, name, frameId)
	local newList = {
		frameId = frameId,
		listType = listType,
		name = name,
		checked = false,
		expanded = true,
		visible = true,
		nextReset = CalculateNextResetTime(listType),
		entries = {}
	}
	table.insert(lists, newList)
	
	return newList
end

function AddEntry(list, name)
	local newEntry = {
		name = name,
		checked = false
	}
	table.insert(list.entries, newEntry)
	
	return newEntry
end

function MigrateOptions(saved, defaults)
	for k, v in pairs(defaults) do
		if(saved[k] == nil) then
			--this is a new option. add the default
			saved[k] = v
		elseif(type(saved[k]) == "table") then
			--the saved option is itself a table, so recurse on it
			MigrateOptions(saved[k], v)
		elseif(saved[k] ~= v) then
			saved[k] = v
		end
	end
end

function GetFrameById(frameId)
	for i, v in ipairs(_frames) do
		if((frameId == nil and v.settings == nil) or (v.settings.frameId == frameId)) then
			return v
		end
	end
	return nil
end

function GetCheckboxHeight(checkbox)
	return math.max(checkbox:GetHeight(), getglobal(checkbox:GetName() .. 'Text'):GetHeight() + 3)
end

function ToggleLocked()
	if(Coylist_CharData.locked) then
		Coylist_CharData.locked = false
		for i, v in ipairs(_frames) do
			v.resizer:Show()
			
			for bi, bv in ipairs(v.controlButtons) do
				bv:Show()
			end
		end
	else
		Coylist_CharData.locked = true
		for i, v in ipairs(_frames) do
			v.resizer:Hide()
			
			for bi, bv in ipairs(v.controlButtons) do
				bv:Hide()
			end
		end
	end
end

function CreateFontString(parent, font, text)
	local fontString = parent:CreateFontString(nil, "ARTWORK")
	fontString:SetJustifyH("LEFT")
	fontString:SetJustifyV("TOP")
	fontString:SetFontObject(font)
	fontString:SetText(text)
	return fontString
end

function ShowColorPicker(r, g, b, a, changedCallback)
	ColorPickerFrame:SetColorRGB(r,g,b);
	ColorPickerFrame.hasOpacity, ColorPickerFrame.opacity = (a ~= nil), a;
	ColorPickerFrame.previousValues = {r,g,b,a};
	ColorPickerFrame.func, ColorPickerFrame.opacityFunc, ColorPickerFrame.cancelFunc = 
	changedCallback, changedCallback, changedCallback;
	ColorPickerFrame:Hide(); -- Need to run the OnShow handler.
	ColorPickerFrame:Show();
end

function DeepCopy(orig)
    local orig_type = type(orig)
    local copy
    if orig_type == 'table' then
        copy = {}
        for orig_key, orig_value in next, orig, nil do
            copy[DeepCopy(orig_key)] = DeepCopy(orig_value)
        end
        setmetatable(copy, DeepCopy(getmetatable(orig)))
    else -- number, string, boolean, etc
        copy = orig
    end
    return copy
end